#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>   // нужно для usbdrv.h
#include "pinmacro.h"
#include "usbdrv.h"
#include "uart.h"
#include "commands.h"

#define VERSION_DEV 2

//команды для e24
#define E24_FREQ 2457600 //тактовая частота внутреннего контроллера
#define E24_ADC_CODE(x) (((E24_FREQ/64) / (x) + 1)>>1) //перевод частоты (Гц) канала АЦП в код лоя e24
#define E24CMD_ENABLE   0x80 //включить канал(ы)
#define E24CMD_INP_SEL  0x90 //управление мультиплексором каналов
#define E24CMD_ADCH     0xA0 //старший байт частоты АЦП
#define E24CMD_ADCL     0xB0 //младший байт частоты АЦП
#define E24CMD_AMPL_CAL 0xC0 //настройка калибровки и диапазона измерения
#define E24CMD_APPLY    0xD0 //применение указанных предыдущими командами параметров
#define E24CMD_COM      0xE0 //(не используется) установка частоты обмена по RS232

//вывод включения питания
#define PWR D,7,0 //в данном случае, PD7, активный уровень - низкий

//флаг произошло ли новое измерение данного канала АЦП с момента последнего опроса
#define ADCF_VALID (1<<0) 

#define ADCS_AMPL   0x07 //маска битов, относящихся к диапазону измерения (амплитуде)
#define ADCS_CAL    0x70 //маска битов, относящихся к калибровке

//флаги режима работы переходника:
#define F_RAW_UART  (1<<0) //передача команд напрямую в e24
#define F_REINIT    (1<<1) //запрос на повторную инициализацию e24
#define F_VER       (1<<2) //запрос версии прошивки

//буфер для результатов АЦП (байт cmd отвечает за настройки, флаги и прочее, байты data0-data2 - значение АЦП)
struct dataexchange_t adc[4] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
uint8_t adc_settings[4] = {0x10, 0x10, 0x10, 0x10}; //настройки, готовящиеся к изменению
uint8_t adc_oldsettings[4] = {0x00, 0x00, 0x00, 0x00}; //настройки, измененииые на самом деле
uint8_t flags = F_REINIT;

struct dataexchange_t pdata = {0, 0, 0, 0};

PROGMEM const char usbHidReportDescriptor[22] = { // USB report descriptor         // Дескриптор описывает структуру пакета данных для обмена
    0x06, 0x00, 0xff,                       // USAGE_PAGE (Generic Desktop)
    0x09, 0x01,                             // USAGE (Vendor Usage 1)
    0xa1, 0x01,                             // COLLECTION (Application)
    0x15, 0x00,                             //    LOGICAL_MINIMUM (0)        // min. значение для данных
    0x26, 0xff, 0x00,                       //    LOGICAL_MAXIMUM (255)      // max. значение для данных, 255 тут не случайно, а чтобы уложиться в 1 байт
    0x75, 0x08,                             //    REPORT_SIZE (8)            // информация передается порциями, это размер одного "репорта" 8 бит
    0x95, sizeof(struct dataexchange_t),    //    REPORT_COUNT               // количество порций
    0x09, 0x00,                             //    USAGE (Undefined)
    0xb2, 0x02, 0x01,                       //    FEATURE (Data,Var,Abs,Buf)
    0xc0                                    // END_COLLECTION
};


// Эти переменные хранят статус текущей передачи
static uchar    currentAddress;
static uchar    bytesRemaining;

/* usbFunctionRead() вызывается когда хост запрашивает порцию данных от устройства
 * Для дополнительной информации см. документацию в usbdrv.h
 */
uchar   usbFunctionRead(uchar *data, uchar len){
  if(len > bytesRemaining)len = bytesRemaining;

  uchar *buffer = (uchar*)&pdata;
  static unsigned char adc_num=0; //номер передаваемого канала АЦП
  unsigned char i, adc_flag;

  if(!currentAddress){
    //была ли последняя команда запросом версии?
    if(flags & F_VER){
      pdata.cmd = REP_VER;
      pdata.data2 = VERSION_DEV; //версия прошивки
      pdata.data1 = VERSION_CMDH; //версия системы команд (ст. байт)
      pdata.data0 = VERSION_CMDL; //версия системы команд (мл. байт)
      flags &=~ F_VER;
    //находится ли переходник в режиме моста USB-RS232 ?
    }else if(flags & F_RAW_UART){
      i = 0; //сколько байт принято по RS232 на самом деле
      if(!uart_is_empty()){i++; pdata.data0 = uart_read();}
      if(!uart_is_empty()){i++; pdata.data1 = uart_read();}
      if(!uart_is_empty()){i++; pdata.data2 = uart_read();}
      pdata.cmd = REP_RAW_UART | i;
    }else{ //штатный режим работы
      i = adc_num;
      do{
        i = (i+1)&0x03; //переход к следующему каналу АЦП
        adc_flag = adc[i].cmd;
        if(adc_flag & ADCF_VALID)break; //если выбранный канал обновился с момента последнего опроса - прекращаем поиск
      }while(i!=adc_num);
      if(adc_flag & ADCF_VALID){ //если выбранный канал обновлялся - отправляем его
        adc_num = i;
        pdata.cmd = REP_ADC | adc_num | ((adc_oldsettings[i] & 0x07)<<2);
        pdata.data2 = adc[i].data2;
        pdata.data1 = adc[i].data1;
        pdata.data0 = adc[i].data0;
        adc[i].cmd &=~ ADCF_VALID;
      }else{ //если ни один канал не обновился - высылаем состояние питания (не NOP же отсылать)
        pdata.cmd = REP_PWR;
        pdata.data2 = 0;
        pdata.data1 = 0;
        pdata.data0 = PIN_ON(PWR);
      }
    }
  }

  uchar j;
  for(j=0; j<len; j++)data[j] = buffer[j+currentAddress];

  currentAddress += len;
  bytesRemaining -= len;
  return len;
}
/* Передача в e24 одного байта
 */
inline void e24_send1(uint8_t cmd){
  uart_write(cmd);
}
/* Передача в e24 двух байт (учитывая особенность e24, на самом деле надо передавать 3 байта)
 */
void e24_send2(uint8_t cmd, uint8_t data){
  uart_write((data>>4) & 0x0F);
  uart_write(data & 0x0F);
  uart_write(cmd);
}

//повторная инициализация e24, реализованная как конечный автомат
#define E24S_MASK   0b11100000
#define E24S_RDY    0b00000000
#define E24S_PWRON  0b00100000
#define E24S_CMDS   0b01000000
void e24_reinit(){
  static unsigned char state = E24S_RDY;
  switch(state & E24S_MASK){
    case E24S_RDY: //выключить e24, настроить UART и выждать пока прибор не перестанет передавать данные
      uart_speed(19200);
      PORT_OFF(PWR);
      while(!uart_is_empty())uart_read();
      _delay_ms(10);
      state++; //это состояние будет повторятся 0b0001111 = 31 раз
      break;
    case E24S_PWRON: //включить e24 и выждать пока не прекратятся переходные процессы в UART
      PORT_ON(PWR);
      state++;
      if(!uart_is_empty())state = E24S_CMDS;
      _delay_ms(10);
      break;
    case E24S_CMDS: //считаем что e24 включился и готов к приему команд
      e24_send2(E24CMD_INP_SEL | 0x0F, 0);
      e24_send2(E24CMD_ADCH | 0x0F, ((E24_ADC_CODE(5))>>8)&0xFF);
      e24_send2(E24CMD_ADCL | 0x0F, (E24_ADC_CODE(5)&0xFF));
      e24_send2(E24CMD_AMPL_CAL | 0x0F, 0x10);
      e24_send1(E24CMD_APPLY | 0x0F);
      e24_send1(E24CMD_ENABLE | 0x0F);
      state = E24S_RDY; //если все хорошо, переходим в состояние "e24 в рабочем режиме"
      flags &=~ F_REINIT; //и "повторная инициализация не требуется"
      break;
    default:break; //что-то пошло не так (в штатном режиме сюда попадать не должны, но вдруг)
      state = E24S_RDY; //включаем повторную инициализацию e24
      flags |= F_REINIT;
  }
}
/* Установка частоты выбранного (выбранных) по маске канала
 */
inline void cmd_adc_freq(uint8_t chan, uint8_t adch, uint8_t adcl){
  uint16_t freq = (adch<<8) | adcl;
  if(freq < 5 || freq > 1037)return;
  freq = E24_ADC_CODE(freq);
  e24_send2(E24CMD_ADCH | chan, (uint8_t)(freq>>8));
  e24_send2(E24CMD_ADCL | chan, (uint8_t)(freq & 0xFF));
}
/* Установка диапазона (максимальной амплитуды)
 */
inline void cmd_ampl(uint8_t chan, uint8_t ampl){
  ampl &= ADCS_AMPL;
  if(chan & 0x01)adc_settings[0] = (adc_settings[0] &~ ADCS_AMPL)|ampl;
  if(chan & 0x02)adc_settings[1] = (adc_settings[1] &~ ADCS_AMPL)|ampl;
  if(chan & 0x04)adc_settings[2] = (adc_settings[2] &~ ADCS_AMPL)|ampl;
  if(chan & 0x08)adc_settings[3] = (adc_settings[3] &~ ADCS_AMPL)|ampl;
}
/* Установка калибровки
 */
inline void cmd_cal(uint8_t chan, uint8_t cal){
  cal &= ADCS_CAL;
  if(chan & 0x01)adc_settings[0] = (adc_settings[0] &~ ADCS_CAL)|cal;
  if(chan & 0x02)adc_settings[1] = (adc_settings[1] &~ ADCS_CAL)|cal;
  if(chan & 0x04)adc_settings[2] = (adc_settings[2] &~ ADCS_CAL)|cal;
  if(chan & 0x08)adc_settings[3] = (adc_settings[3] &~ ADCS_CAL)|cal;
}
/* Настройка мультиплексора
 */
inline void cmd_subchan(uint8_t chan, uint8_t subchan){
  subchan &= 0x03;
  uart_write(0x00); uart_write(subchan); uart_write(E24CMD_INP_SEL | chan);
}
/* Применение сделанных ранее настроек
 */
inline void cmd_apply(uint8_t chan){
  chan &= 0x0F;
  uint8_t i, mask=(1<<0);
  for(i=0;i<4;i++){
    if(chan & mask)if(adc_settings[i] != adc_oldsettings[i]){
      e24_send2(E24CMD_AMPL_CAL | mask, adc_settings[i]);
      adc_oldsettings[i] = adc_settings[i];
    }
    mask <<= 1;
  }
  e24_send1(E24CMD_APPLY | chan);
}
/* Разрешение и запрет работы каналов по маскам (разрешение более приоритетно)
 */
inline void cmd_ch_enable(uint8_t mask_on, uint8_t mask_off){
  mask_on &=~mask_off;
  mask_on &= 0x0F;
  e24_send1(E24CMD_ENABLE | mask_on);
}
/* usbFunctionWrite() вызывается когда хост отправляет порцию данных к устройству
 * Для дополнительной информации см. документацию в usbdrv.h
 */
uchar   usbFunctionWrite(uchar *data, uchar len){
 if(bytesRemaining == 0)return 1;               /* конец передачи */

 if(len > bytesRemaining)len = bytesRemaining;

 uchar *buffer = (uchar*)&pdata;
    
 uchar j;
 for(j=0; j<len; j++)buffer[j+currentAddress] = data[j];

 currentAddress += len;
 bytesRemaining -= len;

 if(bytesRemaining == 0){     // ???<=pdata
   if(flags & F_REINIT)pdata.cmd = CMD_NOP;
   switch(pdata.cmd){
     case CMD_NOP: break;
     case CMD_RAW_UART: flags |= F_RAW_UART; uart_write(pdata.data0); break;
     case CMD_PWR: if(pdata.data0)PORT_ON(PWR); else PORT_OFF(PWR); break;
     case CMD_ADC_FREQ: flags &=~F_RAW_UART; cmd_adc_freq(pdata.data2, pdata.data1, pdata.data0); break;
     case CMD_AMPL: flags &=~F_RAW_UART; cmd_ampl(pdata.data2, pdata.data0); break;
     case CMD_CAL: flags &=~F_RAW_UART; cmd_cal(pdata.data2, pdata.data0); break;
     case CMD_SUBCHANNEL: flags &=~F_RAW_UART; cmd_subchan(pdata.data2, pdata.data0); break;
     case CMD_APPLY: flags &=~F_RAW_UART; cmd_apply(pdata.data2); break;
     case CMD_CH_ENABLE: flags &=~F_RAW_UART; cmd_ch_enable(pdata.data1, pdata.data0); break;
     case CMD_UART_SPEED: uart_speed(((uint16_t)pdata.data1<<8)|(uint16_t)pdata.data0); break;
     case CMD_REINIT: flags &=~F_RAW_UART; flags |= F_REINIT; break;
     case CMD_VER: flags |= F_VER; break;
   }
 }
 return bytesRemaining == 0; /* 0 означает, что есть еще данные */
}

/* ------------------------------------------------------------------------- */

usbMsgLen_t usbFunctionSetup(uchar data[8]){
 usbRequest_t    *rq = (void *)data;

 if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* HID устройство */
  if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
// у нас только одна разновидность репорта, можем игнорировать report-ID
   bytesRemaining = sizeof(struct dataexchange_t);
   currentAddress = 0;
   return USB_NO_MSG;  // используем usbFunctionRead() для отправки данных хосту
  }else if(rq->bRequest == USBRQ_HID_SET_REPORT){
// у нас только одна разновидность репорта, можем игнорировать report-ID
   bytesRemaining = sizeof(struct dataexchange_t);
   currentAddress = 0;
   return USB_NO_MSG;  // используем usbFunctionWrite() для получения данных от хоста
  }
 }else{
/* остальные запросы мы просто игнорируем */
 }
 return 0;
}
/* ------------------------------------------------------------------------- */

inline void adc_update(){
  unsigned char uart_num = uart_rx_size();
  unsigned char uart_cmd, uart_data0, uart_data1, uart_data2;
  if(uart_num == 0)return; //если на UART ничего не пришло - выход
  uart_cmd = uart_scan(); //просматриваем первый пришедший байт (без удаления из очереди)
  if(uart_cmd & 0x80){ //единица в старшем бите - признак начала данных АЦП
    if(uart_num > (UART_SIZE-4))return; //если суммарно пришло менее 4 байт - ждем
    //в противном случае считываем все 4 байта (с удалением из очереди)
    uart_num = uart_read();
    uart_cmd = uart_num;
    uart_data2 = uart_read();
    uart_data1 = uart_read();
    uart_data0 = uart_read();
    //немного черной магии
    asm volatile(
      "andi %0, 0x0F \n\t"
      "lsl %1 \n\t"
      "sbrc %2,6 \n\t"
      "  ori %1,0x01 \n\t"
      "mov r23, %1 \n\t"
      "andi %1, 0xF0 \n\t"
      "or %1, %0 \n\t"
      "swap %1 \n\t"
      "andi r23, 0x0F \n\t"
      "swap r23 \n\t"
      "lsl %3 \n\t"
      "lsr %2 \n\t"
      "ror %3 \n\t"
      "lsr %2 \n\t"
      "ror %3 \n\t"
      "andi %2, 0x0F \n\t"
      "or %2, r23 \n\t"
      : "+a"(uart_cmd), "+a"(uart_data2), "+a"(uart_data1), "+a"(uart_data0)
      :
      : "r23","memory"
    );
    //на самом деле - преобразование АЦП из формата e24 в обычное 24-битное число
    uart_num = (uart_num >> 4)&0x03;
    adc[uart_num].cmd = ADCF_VALID;
    adc[uart_num].data2 = uart_data2;
    adc[uart_num].data1 = uart_data1;
    adc[uart_num].data0 = uart_data0;
  }else{
    //если пришло что-то неожиданное - игнорируем
    uart_read();
  }
}

int main(void){
 cli();
 uart_init();
 
 usbInit();
 usbDeviceDisconnect();  // принудительно отключаемся от хоста, так делать можно только при выключенных прерываниях!

 DDR_1(PWR);
 PORT_ON(PWR);
 
 _delay_ms(255);
    
 usbDeviceConnect();     // подключаемся
 sei();                  // разрешаем прерывания

 for(;;){                // главный цикл программы
  usbPoll();          // эту функцию надо регулярно вызывать с главного цикла, максимальная задержка между вызовами - 50 ms
  uart_update();
  if(flags & F_REINIT){
    e24_reinit();
  }else if(!(flags & F_RAW_UART)){
    adc_update();
  }
  
 }
 return 0;
}
/* ------------------------------------------------------------------------- */ 
